classdef class_pair_elem_handle < handle
  properties
    value
  endproperties

  methods

    function this = class_pair_elem_handle (value)
      this.value = value;
    endfunction

  endmethods
endclassdef
