// DO NOT EDIT -- generated by mk-ops.awk
#if ! defined (octave_mx_fnda_i16nda_h)
#define octave_mx_fnda_i16nda_h 1
#include "octave-config.h"
#include "int16NDArray.h"
#include "fNDArray.h"
  extern OCTAVE_API int16NDArray operator + (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API int16NDArray operator - (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API int16NDArray product (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API int16NDArray quotient (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_lt (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_le (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_ge (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_gt (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_eq (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_ne (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_and (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_or (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_not_and (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_not_or (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_and_not (const FloatNDArray&, const int16NDArray&);
  extern OCTAVE_API boolNDArray mx_el_or_not (const FloatNDArray&, const int16NDArray&);
#endif
